# treetable.tcl --
#
# This file defines the default bindings for Tk listbox widgets
# and provides procedures that help in implementing those bindings.
#
# $Id: treetable.tcl 1.1 1996/03/17 04:19:22 heller Exp $
#
# Copyright (c) 1994 The Regents of the University of California.
# Copyright (c) 1994-1995 Sun Microsystems, Inc.
#
# See the file "license.terms" for information on usage and redistribution
# of this file, and for a DISCLAIMER OF ALL WARRANTIES.
#
# Adapted for the TreeTable class
#--------------------------------------------------------------------------
# tkPriv elements used in this file:
#
# afterId -		Token returned by "after" for autoscanning.
# treetablePrev -		The last element to be selected or deselected
#			during a selection operation.
# treetableSelection -	All of the items that were selected before the
#			current selection operation (such as a mouse
#			drag) started;  used to cancel an operation.
#--------------------------------------------------------------------------
# tkScreenChanged --
# This procedure is invoked by the binding mechanism whenever the
# "current" screen is changing.  The procedure does two things.
# First, it uses "upvar" to make global variable "tkPriv" point at an
# array variable that holds state for the current display.  Second,
# it initializes the array if it didn't already exist.
#
# Arguments:
# screen -		The name of the new screen.

proc tkScreenChanged screen {
    set disp [file rootname $screen]
    uplevel #0 upvar #0 tkPriv.$disp tkPriv
    global tkPriv
    if [info exists tkPriv] {
	set tkPriv(screen) $screen
	return
    }
    set tkPriv(afterId) {}
    set tkPriv(buttons) 0
    set tkPriv(buttonWindow) {}
    set tkPriv(dragging) {}
    set tkPriv(focus) {}
    set tkPriv(grab) {}
    set tkPriv(initPos) {}
    set tkPriv(inMenubutton) {}
    set tkPriv(listboxPrev) {}
# Added for treetables RPH
    set tkPriv(treetablePrev) {}
    set tkPriv(mouseMoved) 0
    set tkPriv(popup) {}
    set tkPriv(postedMb) {}
    set tkPriv(pressX) 0
    set tkPriv(pressY) 0
    set tkPriv(screen) $screen
    set tkPriv(selectMode) char
    set tkPriv(window) {}
}

set tkPriv(treetablePrev) {}


#-------------------------------------------------------------------------
# The code below creates the default class bindings for treetablees.
#-------------------------------------------------------------------------

bind TreeTable <1> {
    tkTreeTableBeginSelect %W [%W index @%x,%y]
}
bind TreeTable <B1-Motion> {
    set tkPriv(x) %x
    set tkPriv(y) %y
    tkTreeTableMotion %W [%W index @%x,%y]
}
bind TreeTable <ButtonRelease-1> {
    tkCancelRepeat
    %W activate @%x,%y
}
bind TreeTable <Shift-1> {
    tkTreeTableBeginExtend %W [%W index @%x,%y]
}
bind TreeTable <Control-1> {
    tkTreeTableBeginToggle %W [%W index @%x,%y]
}
bind TreeTable <B1-Leave> {
    set tkPriv(x) %x
    set tkPriv(y) %y
    tkTreeTableAutoScan %W
}
bind TreeTable <B1-Enter> {
    tkCancelRepeat
}

bind TreeTable <Up> {
    tkTreeTableUpDown %W -1
}
bind TreeTable <Shift-Up> {
    tkTreeTableExtendUpDown %W -1
}
bind TreeTable <Down> {
    tkTreeTableUpDown %W 1
}
bind TreeTable <Shift-Down> {
    tkTreeTableExtendUpDown %W 1
}
bind TreeTable <Left> {
    %W xview scroll -1 units
}
bind TreeTable <Control-Left> {
    %W xview scroll -1 pages
}
bind TreeTable <Right> {
    %W xview scroll 1 units
}
bind TreeTable <Control-Right> {
    %W xview scroll 1 pages
}
bind TreeTable <Prior> {
    %W yview scroll -1 pages
}
bind TreeTable <Next> {
    %W yview scroll 1 pages
}
bind TreeTable <Control-Prior> {
    %W xview scroll -1 pages
}
bind TreeTable <Control-Next> {
    %W xview scroll 1 pages
}
bind TreeTable <Home> {
    %W xview moveto 0
}
bind TreeTable <End> {
    %W xview moveto 1
}
bind TreeTable <Control-Home> {
    %W activate 0
    %W see 0
    %W selection clear 0 end
    %W selection set 0
}
bind TreeTable <Shift-Control-Home> {
    tkTreeTableDataExtend %W 0
}
bind TreeTable <Control-End> {
    %W activate end
    %W see end
    %W selection clear 0 end
    %W selection set end
}
bind TreeTable <Shift-Control-End> {
    tkTreeTableDataExtend %W end
}
bind TreeTable <F16> {
    if {[selection own -displayof %W] == "%W"} {
	clipboard clear -displayof %W
	clipboard append -displayof %W [selection get -displayof %W]
    }
}
bind TreeTable <space> {
    tkTreeTableBeginSelect %W [%W index active]
}
bind TreeTable <Select> {
    tkTreeTableBeginSelect %W [%W index active]
}
bind TreeTable <Control-Shift-space> {
    tkTreeTableBeginExtend %W [%W index active]
}
bind TreeTable <Shift-Select> {
    tkTreeTableBeginExtend %W [%W index active]
}
bind TreeTable <Escape> {
    tkTreeTableCancel %W
}
bind TreeTable <Control-slash> {
    tkTreeTableSelectAll %W
}
bind TreeTable <Control-backslash> {
    if {[%W cget -selectmode] != "browse"} {
	%W selection clear 0 end
    }
}

# Additional Tk bindings that aren't part of the Motif look and feel:

bind TreeTable <2> {
    %W scan mark %x %y
}
bind TreeTable <B2-Motion> {
    %W scan dragto %x %y
}

# tkTreeTableBeginSelect --
#
# This procedure is typically invoked on button-1 presses.  It begins
# the process of making a selection in the treetable.  Its exact behavior
# depends on the selection mode currently in effect for the treetable;
# see the Motif documentation for details.
#
# Arguments:
# w -		The treetable widget.
# el -		The element for the selection operation (typically the
#		one under the pointer).  Must be in numerical form.

proc tkTreeTableBeginSelect {w el} {
    global tkPriv
    if {[$w cget -selectmode]  == "multiple"} {
	if [$w selection includes $el] {
	    $w selection clear $el
	} else {
	    $w selection set $el
	}
    } else {
	$w selection clear 0 end
	$w selection set $el
	$w selection anchor $el
	set tkPriv(treetableSelection) {}
	set tkPriv(treetablePrev) $el
    }
}

# tkTreeTableMotion --
#
# This procedure is called to process mouse motion events while
# button 1 is down.  It may move or extend the selection, depending
# on the treetable's selection mode.
#
# Arguments:
# w -		The treetable widget.
# el -		The element under the pointer (must be a number).

proc tkTreeTableMotion {w el} {
    global tkPriv
    if {$el == $tkPriv(treetablePrev)} {
	return
    }
    set anchor [$w index anchor]
    switch [$w cget -selectmode] {
	browse {
	    $w selection clear 0 end
	    $w selection set $el
	    set tkPriv(treetablePrev) $el
	}
	extended {
	    set i $tkPriv(treetablePrev)
	    if [$w selection includes anchor] {
		$w selection clear $i $el
		$w selection set anchor $el
	    } else {
		$w selection clear $i $el
		$w selection clear anchor $el
	    }
	    while {($i < $el) && ($i < $anchor)} {
		if {[lsearch $tkPriv(treetableSelection) $i] >= 0} {
		    $w selection set $i
		}
		incr i
	    }
	    while {($i > $el) && ($i > $anchor)} {
		if {[lsearch $tkPriv(treetableSelection) $i] >= 0} {
		    $w selection set $i
		}
		incr i -1
	    }
	    set tkPriv(treetablePrev) $el
	}
    }
}

# tkTreeTableBeginExtend --
#
# This procedure is typically invoked on shift-button-1 presses.  It
# begins the process of extending a selection in the treetable.  Its
# exact behavior depends on the selection mode currently in effect
# for the treetable;  see the Motif documentation for details.
#
# Arguments:
# w -		The treetable widget.
# el -		The element for the selection operation (typically the
#		one under the pointer).  Must be in numerical form.

proc tkTreeTableBeginExtend {w el} {
    if {([$w cget -selectmode] == "extended")
	    && [$w selection includes anchor]} {
	tkTreeTableMotion $w $el
    }
}

# tkTreeTableBeginToggle --
#
# This procedure is typically invoked on control-button-1 presses.  It
# begins the process of toggling a selection in the treetable.  Its
# exact behavior depends on the selection mode currently in effect
# for the treetable;  see the Motif documentation for details.
#
# Arguments:
# w -		The treetable widget.
# el -		The element for the selection operation (typically the
#		one under the pointer).  Must be in numerical form.

proc tkTreeTableBeginToggle {w el} {
    global tkPriv
    if {[$w cget -selectmode] == "extended"} {
	set tkPriv(treetableSelection) [$w curselection]
	set tkPriv(treetablePrev) $el
	$w selection anchor $el
	if [$w selection includes $el] {
	    $w selection clear $el
	} else {
	    $w selection set $el
	}
    }
}

# tkTreeTableAutoScan --
# This procedure is invoked when the mouse leaves an entry window
# with button 1 down.  It scrolls the window up, down, left, or
# right, depending on where the mouse left the window, and reschedules
# itself as an "after" command so that the window continues to scroll until
# the mouse moves back into the window or the mouse button is released.
#
# Arguments:
# w -		The entry window.

proc tkTreeTableAutoScan {w} {
    global tkPriv
    set x $tkPriv(x)
    set y $tkPriv(y)
    if {$y >= [winfo height $w]} {
	$w yview scroll 1 units
    } elseif {$y < 0} {
	$w yview scroll -1 units
    } elseif {$x >= [winfo width $w]} {
	$w xview scroll 2 units
    } elseif {$x < 0} {
	$w xview scroll -2 units
    } else {
	return
    }
    tkTreeTableMotion $w [$w index @$x,$y]
    set tkPriv(afterId) [after 50 tkTreeTableAutoScan $w]
}

# tkTreeTableUpDown --
#
# Moves the location cursor (active element) up or down by one element,
# and changes the selection if we're in browse or extended selection
# mode.
#
# Arguments:
# w -		The treetable widget.
# amount -	+1 to move down one item, -1 to move back one item.

proc tkTreeTableUpDown {w amount} {
    global tkPriv
    $w activate [expr [$w index active] + $amount]
    $w see active
    switch [$w cget -selectmode] {
	browse {
	    $w selection clear 0 end
	    $w selection set active
	}
	extended {
	    $w selection clear 0 end
	    $w selection set active
	    $w selection anchor active
	    set tkPriv(treetablePrev) [$w index active]
	    set tkPriv(treetableSelection) {}
	}
    }
}

# tkTreeTableExtendUpDown --
#
# Does nothing unless we're in extended selection mode;  in this
# case it moves the location cursor (active element) up or down by
# one element, and extends the selection to that point.
#
# Arguments:
# w -		The treetable widget.
# amount -	+1 to move down one item, -1 to move back one item.

proc tkTreeTableExtendUpDown {w amount} {
    if {[$w cget -selectmode] != "extended"} {
	return
    }
    $w activate [expr [$w index active] + $amount]
    $w see active
    tkTreeTableMotion $w [$w index active]
}

# tkTreeTableDataExtend
#
# This procedure is called for key-presses such as Shift-KEndData.
# If the selection mode isn't multiple or extend then it does nothing.
# Otherwise it moves the active element to el and, if we're in
# extended mode, extends the selection to that point.
#
# Arguments:
# w -		The treetable widget.
# el -		An integer element number.

proc tkTreeTableDataExtend {w el} {
    set mode [$w cget -selectmode]
    if {$mode == "extended"} {
	$w activate $el
	$w see $el
        if [$w selection includes anchor] {
	    tkTreeTableMotion $w $el
	}
    } elseif {$mode == "multiple"} {
	$w activate $el
	$w see $el
    }
}

# tkTreeTableCancel
#
# This procedure is invoked to cancel an extended selection in
# progress.  If there is an extended selection in progress, it
# restores all of the items between the active one and the anchor
# to their previous selection state.
#
# Arguments:
# w -		The treetable widget.

proc tkTreeTableCancel w {
    global tkPriv
    if {[$w cget -selectmode] != "extended"} {
	return
    }
    set first [$w index anchor]
    set last $tkPriv(treetablePrev)
    if {$first > $last} {
	set tmp $first
	set first $last
	set last $tmp
    }
    $w selection clear $first $last
    while {$first <= $last} {
	if {[lsearch $tkPriv(treetableSelection) $first] >= 0} {
	    $w selection set $first
	}
	incr first
    }
}

# tkTreeTableSelectAll
#
# This procedure is invoked to handle the "select all" operation.
# For single and browse mode, it just selects the active element.
# Otherwise it selects everything in the widget.
#
# Arguments:
# w -		The treetable widget.

proc tkTreeTableSelectAll w {
    set mode [$w cget -selectmode]
    if {($mode == "single") || ($mode == "browse")} {
	$w selection clear 0 end
	$w selection set active
    } else {
	$w selection set 0 end
    }
}
